/*
 * Device driver for the SYMBIOS/LSILOGIC 53C8XX and 53C1010 family 
 * of PCI-SCSI IO processors.
 *
 * Copyright (C) 1999-2001  Gerard Roudier <groudier@club-internet.fr>
 *
 * This driver is derived from the Linux sym53c8xx driver.
 * Copyright (C) 1998-2000  Gerard Roudier
 *
 * The sym53c8xx driver is derived from the ncr53c8xx driver that had been 
 * a port of the FreeBSD ncr driver to Linux-1.2.13.
 *
 * The original ncr driver has been written for 386bsd and FreeBSD by
 *         Wolfgang Stanglmeier        <wolf@cologne.de>
 *         Stefan Esser                <se@mi.Uni-Koeln.de>
 * Copyright (C) 1994  Wolfgang Stanglmeier
 *
 * Other major contributions:
 *
 * NVRAM detection and reading.
 * Copyright (C) 1997 Richard Waltham <dormouse@farsrobt.demon.co.uk>
 *
 *-----------------------------------------------------------------------------
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * Where this Software is combined with software released under the terms of 
 * the GNU Public License ("GPL") and the terms of the GPL would require the 
 * combined work to also be released under the terms of the GPL, the terms
 * and conditions of this License will apply in addition to those of the
 * GPL with the exception of any terms or conditions of this License that
 * conflict with, or are expressly prohibited by, the GPL.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef SYM_GLUE_H
#define SYM_GLUE_H

#include <sys/param.h>

#if	__NetBSD_Version__ < 104000200
#error	"This driver requires a kernel version not lower than 1.4.2"
#endif

#if	__NetBSD_Version__ > 999999999
#error	"This driver requires a kernel version not greater than 99.9999.999 :)"
#endif

#include <sys/types.h>
#include <sys/systm.h>
#if __NetBSD_Version__ >= 105000000
#include <sys/callout.h>
#endif
#include <sys/kernel.h>
#if __NetBSD_Version__ >= 106000000
#include <sys/reboot.h>
#endif
#include <sys/errno.h>  
#include <sys/ioctl.h>
#include <sys/scsiio.h>
#include <sys/device.h>
#include <sys/malloc.h>
#include <sys/buf.h> 
#include <sys/proc.h>
#include <sys/user.h>

#include <dev/scsipi/scsi_all.h>
#include <dev/scsipi/scsipi_all.h>
#include <dev/scsipi/scsiconf.h>
#include <dev/scsipi/scsi_message.h>
#include <dev/scsipi/scsipi_debug.h>

#if __NetBSD_Version__ >= 106000000
#include <uvm/uvm_extern.h>
#else
#include <vm/vm.h>
#include <vm/vm_param.h>
#include <vm/pmap.h>
#endif

#include <machine/bswap.h>
#include <machine/bus.h>
#include <machine/intr.h>

#include <dev/pci/pcireg.h>
#include <dev/pci/pcivar.h>

/* #include "opt_sym.h" */

/*
 *  For now, the driver memory allocator is expecting a 
 *  PAGE size value known at compile time.
 *  Shoe-horn some safe value assumed to be smaller than 
 *  the actual PAGE size. This will be cleaned up later.
 */
#ifdef  PGSHIFT
#define SYM_PAGE_SHIFT		PGSHIFT
#else
#define SYM_PAGE_SHIFT		12
#endif

/*
 *  Since NetBSD does not handle a per SCSI BUS queue depth,
 *  it seems desirable to use the greatest possible 
 *  number of outstanding command per SCSI BUS and some 
 *  reasonnable max number of tasks per LUN (32).
 */
#define SYM_CONF_MAX_START	(508)
#define SYM_CONF_MAX_TAG_ORDER	(6)
#define SYM_CONF_MAX_TAG	(32)

/*
 *  More than 8 LUNs seems to make problems.
 *  This will be investigated when time will allow.
 */
#define SYM_CONF_MAX_LUN	(8)

/*
 *  Still a couple of ugly hacks. :)
 */
#define vm_offset_t		bus_addr_t
#define printf_info		printf
#ifndef	__init
#define	__init
#endif
#ifndef	__initdata
#define	__initdata
#endif

/*  Just for testing purpose */
/* #define DEBUG_FLAGS		0x201 */
#define DEBUG_FLAGS		0x000

#if 0
#define	DEBUG_PAUSE(msg)	\
	do {printf("XXXX %s XXXX\n", msg); MDELAY(100);} while (0)
#define	DEBUG_FREEZE(msg)	\
	do {printf("XXXX %s XXXX\n", msg); MDELAY(2000);} while (1)
#endif

/* Short and quite clear integer types */
typedef int8_t    s8;
typedef int16_t   s16;
typedef	int32_t   s32;
typedef u_int8_t  u8;
typedef u_int16_t u16;
typedef	u_int32_t u32;

/*
 *  General driver includes.
 */
#include "sym_misc.h"
#include "sym_conf.h"
#include "sym_defs.h"

/*
 *  Configuration addendum for NetBSD.
 */
/* #define SYM_OPT_HANDLE_DIR_UNKNOWN */
#define SYM_OPT_HANDLE_DEVICE_QUEUEING
/* #define SYM_OPT_NVRAM_PRE_READ */
#define SYM_OPT_SNIFF_INQUIRY
#define SYM_OPT_LIMIT_COMMAND_REORDERING
#define SYM_OPT_HANDLE_IO_TIMEOUT
#define	SYM_OPT_ANNOUNCE_TRANSFER_RATE
#define SYM_OPT_BUS_DMA_ABSTRACTION

/*
 *  Insert a delay in micro-seconds and milli-seconds.
 */
void sym_udelay(int us);
void sym_mdelay(int ms);

/*
 *  Let the compiler know about driver data structure names.
 */
typedef struct sym_tcb *tcb_p;
typedef struct sym_lcb *lcb_p;
typedef struct sym_ccb *ccb_p;
typedef struct sym_hcb *hcb_p;
typedef struct sym_stcb *stcb_p;
typedef struct sym_slcb *slcb_p;
typedef struct sym_sccb *sccb_p;
typedef struct sym_shcb *shcb_p;

/*
 *  Define a reference to the O/S dependant IO request.
 */
typedef struct scsipi_xfer *cam_ccb_p;	 /* Generic */
typedef struct scsipi_xfer *cam_scsiio_p;/* SCSI I/O */

/*
 *  Some provision for a possible big endian mode supported by 
 *  Symbios chips (never seen, by the way).
 *  For now, this stuff does not deserve any comments. :)
 */
#define sym_offb(o)	(o)
#define sym_offw(o)	(o)

/*
 *  Some provision for support for BIG ENDIAN CPU.
 */
#if	BYTE_ORDER == BIG_ENDIAN
#define cpu_to_scr(dw)	_htole32(dw)
#define scr_to_cpu(dw)	_le32toh(dw)
#else
#define cpu_to_scr(dw)	(dw)
#define scr_to_cpu(dw)	(dw)
#endif

/*
 *  Access to the chip IO registers and on-chip RAM.
 */

#if defined(SYM_CONF_IOMAPPED)

#define INB_OFF(o)	bus_space_read_1(np->s.io.tag, np->s.io.bsh, o)
#define INW_OFF(o)	bus_space_read_2(np->s.io.tag, np->s.io.bsh, o)
#define INL_OFF(o)	bus_space_read_4(np->s.io.tag, np->s.io.bsh, o)

#define OUTB_OFF(o, v)	bus_space_write_1(np->s.io.tag, np->s.io.bsh, o, (v))
#define OUTW_OFF(o, v)	bus_space_write_2(np->s.io.tag, np->s.io.bsh, o, (v))
#define OUTL_OFF(o, v)	bus_space_write_4(np->s.io.tag, np->s.io.bsh, o, (v))

#else	/* Memory mapped IO */

#define INB_OFF(o)	bus_space_read_1(np->s.mmio.tag, np->s.mmio.bsh, o)
#define INW_OFF(o)	bus_space_read_2(np->s.mmio.tag, np->s.mmio.bsh, o)
#define INL_OFF(o)	bus_space_read_4(np->s.mmio.tag, np->s.mmio.bsh, o)

#define OUTB_OFF(o, v) bus_space_write_1(np->s.mmio.tag, np->s.mmio.bsh, o, (v))
#define OUTW_OFF(o, v) bus_space_write_2(np->s.mmio.tag, np->s.mmio.bsh, o, (v))
#define OUTL_OFF(o, v) bus_space_write_4(np->s.mmio.tag, np->s.mmio.bsh, o, (v))

#endif	/* SYM_CONF_IOMAPPED */

#define OUTRAM_OFF(o, a, l)	\
	_bus_space_write_region_1_(np->s.ram.tag, np->s.ram.bsh, o, (a), (l))

/*
 *  Using directly the macro does fault, probably due to 
 *  miscompilation. Thus this C function that fixes the problem.
 */
void 
_bus_space_write_region_1_(bus_space_tag_t, bus_space_handle_t, 
			    bus_size_t, u8 *, bus_size_t);
/*
 *  Remap some status field values.
 */
#define CAM_REQ_CMP		XS_NOERROR
#define CAM_SEL_TIMEOUT		XS_SELTIMEOUT
#define CAM_CMD_TIMEOUT		XS_TIMEOUT
#define CAM_REQ_ABORTED		XS_DRIVER_STUFFUP
#define CAM_UNCOR_PARITY	XS_DRIVER_STUFFUP
#define CAM_SCSI_BUS_RESET	XS_RESET
#define CAM_REQUEUE_REQ		0x80
#define	CAM_UNEXP_BUSFREE	XS_DRIVER_STUFFUP
#define	CAM_SCSI_BUSY		XS_BUSY

#define	CAM_DEV_NOT_THERE	XS_DRIVER_STUFFUP
#define	CAM_REQ_INVALID		XS_DRIVER_STUFFUP
#define	CAM_REQ_TOO_BIG		XS_DRIVER_STUFFUP

/*
 *  Remap SCSI data direction values.
 */
#define CAM_DIR_NONE		0
#define CAM_DIR_IN		1
#define CAM_DIR_OUT		2

/*
 *  Misc. constants.
 */
#define SYM_SNS_BBUF_LEN	sizeof(struct scsipi_sense_data)

/*
 *  System specific target data structure.
 *  None for now.
 */
/* #define SYM_HAVE_STCB */

/*
 *  System specific lun data structure.
 */
#define SYM_HAVE_SLCB
struct sym_slcb {
	u_short	reqtags;	/* Number of tags requested by user */
	u_short openings;	/* Last value we set for this device */
};

/*
 *  System specific command data structure.
 */
#define	SYM_HAVE_SCCB
struct sym_sccb {
	bus_dmamap_t	dmamap;		/* DMA map for user data	*/
	u_char		dmamapped;	/* Direction we mapped it for	*/
#define SYM_DMA_NONE	0
#define SYM_DMA_READ	1
#define SYM_DMA_WRITE	2
};

/*
 *  Structures that host the mapping of IO and MMIO resources.
 */
struct sym_bus_mem {
	bus_space_handle_t bsh;
	bus_space_tag_t    tag;
};
struct sym_bus_io {
	bus_space_handle_t bsh;
	bus_space_tag_t    tag;
};

/*
 *  System specific host data structure.
 */
struct sym_shcb {
	/*
	 *  Chip and controller indentification.
	 */
	struct device *device;

	/*
	 *  SIM information for this instance.
	 */
	struct	scsipi_link sc_link;
	struct	scsipi_adapter sc_adapter;

	/*
	 *  Allocated hardware resources.
	 */
	struct sym_bus_mem mmio;
	struct sym_bus_mem ram;
	struct sym_bus_io  io;
	void *intr;

	/*
	 *  DMA stuff.
	 */
	bus_dma_tag_t	data_dmat;	/* Tag for user data DMA	*/

	/*
	 *  To access PCI configuration space.
	 */
	pci_chipset_tag_t pa_pc;
	pcitag_t pa_tag;

	/*
	 *  Miscellaneous ...
	 */
#if __NetBSD_Version__ >= 105000000
	struct callout clock_ch;
#endif
};

/*
 *  Return the name of the controller.
 */
#define	sym_name(np) (np)->s.device->dv_xname

/*
 *  Data structure pointer used as input for the NVRAM reading.
 *  Must resolve the IO macros and sym_name().
 *  The host data structure fits just fine here.
 */
typedef hcb_p sdev_p;

/*
 *  The driver definitions (sym_hipd.h) must know about a 
 *  couple of things related to the memory allocator.
 */
typedef u_long m_addr_t;	/* Enough bits to represent any address */
#define SYM_MEM_PAGE_ORDER 0	/* 1 PAGE  maximum */
#define SYM_MEM_CLUSTER_SHIFT	(SYM_PAGE_SHIFT+SYM_MEM_PAGE_ORDER)
#if 0
#define SYM_MEM_FREE_UNUSED	/* Free unused pages immediately */
#endif

typedef bus_dma_tag_t m_pool_ident_t;

#define SYM_HAVE_M_SVTOB
struct sym_m_svtob {
	bus_dmamap_t dmamap;	/* Dma Map for this chunk */
};

#define SYM_HAVE_M_SPOOL
struct sym_m_spool {
	bus_dma_tag_t dmat;	/* Tag for our fixed allocations */
};

/*
 *  Include driver soft definitions.
 */
#include "sym_fw.h"
#include "sym_hipd.h"

/*
 *  Memory allocator.
 */
#define SYM_MEM_WARN	1	/* Warn on failed operations */

#define sym_get_mem_cluster()	malloc(SYM_MEM_CLUSTER_SIZE, M_DEVBUF, M_NOWAIT)
#define sym_free_mem_cluster(p)	free((p), M_DEVBUF)

#define	sym_calloc(size, name)	 sym_calloc_unlocked(size, name)
#define	sym_mfree(m, size, name) sym_mfree_unlocked(m, size, name)

/*
 *  We have to provide the driver memory allocator with methods for 
 *  it to maintain virtual to bus physical address translations.
 */

#define sym_m_pool_match(mp_id1, mp_id2)	(mp_id1 == mp_id2)

static __inline m_addr_t sym_m_get_dma_mem_cluster(m_pool_p mp, m_vtob_p vbp)
{
	bus_dma_segment_t seg;
	int nseg = 0;
	int err;
	char *errm = "";
	void *vaddr = 0;

	err = bus_dmamem_alloc(mp->s.dmat, SYM_MEM_CLUSTER_SIZE, 
			       SYM_MEM_CLUSTER_SIZE, SYM_MEM_CLUSTER_SIZE,
			       &seg, 1, &nseg, BUS_DMA_NOWAIT);
	if (err) {
		errm = "DMAable memory allocation failed";
		nseg = 0;
		goto out_err;
	}

	err = bus_dmamem_map(mp->s.dmat, &seg, nseg, SYM_MEM_CLUSTER_SIZE,
			(caddr_t *)&vaddr, BUS_DMA_NOWAIT|BUS_DMA_COHERENT);
	if (err) {
		errm = "unable to map DMAable memory";
		vaddr = 0;
		goto out_err;
	}

	err = bus_dmamap_create(mp->s.dmat, SYM_MEM_CLUSTER_SIZE, 1,
			SYM_MEM_CLUSTER_SIZE, SYM_MEM_CLUSTER_SIZE, 
			BUS_DMA_NOWAIT|BUS_DMA_ALLOCNOW, &vbp->s.dmamap);
	if (err) {
                errm = "unable to create DMA map";
		vbp->s.dmamap = 0;
		goto out_err;
        }

	err = bus_dmamap_load(mp->s.dmat, vbp->s.dmamap, vaddr,
			SYM_MEM_CLUSTER_SIZE, 0, BUS_DMA_NOWAIT);
	if (err) {
                errm = "unable to load DMA map";
		goto out_err;
        }

	vbp->baddr = seg.ds_addr;
	vbp->vaddr = (m_addr_t) vaddr;

	return (m_addr_t) vaddr;

out_err:
	printf("sym: error=%d\n", err);
	if (vbp->s.dmamap)
		bus_dmamap_destroy(mp->s.dmat, vbp->s.dmamap);
	if (vaddr)
		bus_dmamem_unmap(mp->s.dmat, vaddr, SYM_MEM_CLUSTER_SIZE);
	if (nseg)
		bus_dmamem_free(mp->s.dmat, &seg, nseg);
	return 0;
}

static __inline void sym_m_free_dma_mem_cluster(m_pool_p mp, m_vtob_p vbp)
{
	bus_dma_segment_t seg;

	seg.ds_addr = vbp->baddr;
	seg.ds_len  = SYM_MEM_CLUSTER_SIZE;

	bus_dmamap_unload(mp->s.dmat, vbp->s.dmamap);
	bus_dmamap_destroy(mp->s.dmat, vbp->s.dmamap);
	bus_dmamem_unmap(mp->s.dmat, (void *) vbp->vaddr,
			 SYM_MEM_CLUSTER_SIZE);
	bus_dmamem_free(mp->s.dmat, &seg, 1);
}

static __inline int sym_m_create_dma_mem_tag(m_pool_p mp)
{
	mp->s.dmat = mp->dev_dmat;
	return 0;
}

static __inline void sym_m_delete_dma_mem_tag(m_pool_p mp)
{
	return;
}

#define __sym_calloc_dma(dev_dmat, size, name)		\
		__sym_calloc_dma_unlocked(dev_dmat, size, name)
#define __sym_mfree_dma(dev_dmat, m, size, name)	\
		__sym_mfree_dma_unlocked(dev_dmat, m, size, name)
#define __vtobus(dev_dmat, m)	\
		__vtobus_unlocked(dev_dmat, m)

/*
 *  Set the status field of a CAM CCB.
 */
static __inline void 
sym_set_cam_status(struct scsipi_xfer *xp, int status)
{
	xp->error = status;
}

/*
 *  Get the status field of a CAM CCB.
 */
static __inline int 
sym_get_cam_status(struct scsipi_xfer *xp)
{
	return xp->error;
}

/*
 *  Create a dma map for this IO.
 */
static __inline int  
sym_data_dmamap_create(hcb_p np, ccb_p cp)
{
	int err;

	err =  bus_dmamap_create(np->s.data_dmat,
			(SYM_CONF_MAX_SG - 1) * PAGE_SIZE, SYM_CONF_MAX_SG,
			(SYM_CONF_MAX_SG - 1) * PAGE_SIZE, 0,
			BUS_DMA_NOWAIT|BUS_DMA_ALLOCNOW,
			&cp->s.dmamap);
	return err;
}
/*
 *  Destroy the dma map of this IO.
 */
static __inline void  
sym_data_dmamap_destroy(hcb_p np, ccb_p cp)
{
	bus_dmamap_destroy(np->s.data_dmat, cp->s.dmamap);
}
/*
 *  Unload the dma map for this IO.
 */
static __inline void 
sym_data_dmamap_unload(hcb_p np, ccb_p cp)
{
	if (cp->s.dmamapped) {
		bus_dmamap_unload(np->s.data_dmat, cp->s.dmamap);
		cp->s.dmamapped = 0;
	}
}
static __inline void 
sym_data_dmamap_presync(hcb_p np, ccb_p cp)
{
	if (cp->s.dmamapped) {
		bus_dmamap_sync(np->s.data_dmat, cp->s.dmamap, 0,
				cp->s.dmamap->dm_mapsize,
				cp->s.dmamapped == SYM_DMA_READ ?
				BUS_DMASYNC_PREREAD : BUS_DMASYNC_PREWRITE);
	}
}
static __inline void 
sym_data_dmamap_postsync(hcb_p np, ccb_p cp)
{
	if (cp->s.dmamapped) {
		bus_dmamap_sync(np->s.data_dmat, cp->s.dmamap, 0,
				cp->s.dmamap->dm_mapsize,
				cp->s.dmamapped == SYM_DMA_READ ?
				BUS_DMASYNC_POSTREAD : BUS_DMASYNC_POSTWRITE);	
	}
}

/*
 *  Async handler for negotiations.
 */
void sym_xpt_async_nego_wide(hcb_p np, int target);
#define sym_xpt_async_nego_sync(np, target)	\
	sym_announce_transfer_rate(np, target)
#define sym_xpt_async_nego_ppr(np, target)	\
	sym_announce_transfer_rate(np, target)

/*
 *  Build CAM result for a successful IO and for a failed IO.
 */
static __inline void sym_set_cam_result_ok(hcb_p np, ccb_p cp, int resid)
{
	struct scsipi_xfer *xp = cp->cam_ccb;

	xp->resid  = resid;
	xp->error  = XS_NOERROR;
	xp->status = cp->ssss_status;

}
void sym_set_cam_result_error(hcb_p np, ccb_p cp, int resid);

/*
 *  We are using the optionnal driver IO timeout handling,
 *  with a clock resolution of 1 second.
 */
#define SYM_CLOCK_TICKS		(hz)	
#define sym_cam_timeout(ccb)	((((ccb)->timeout * hz) + 999)/1000)

/*
 *  Other O/S specific methods.
 */
#define sym_cam_target_id(ccb)	(ccb)->sc_link->scsipi_scsi.target
#define sym_cam_target_lun(ccb)	(ccb)->sc_link->scsipi_scsi.lun

#define	sym_freeze_cam_ccb(ccb)	do { ; } while (0)
void sym_xpt_done(hcb_p np, cam_ccb_p ccb);
void sym_xpt_done2(hcb_p np, cam_ccb_p ccb, int cam_status);
void sym_print_addr (ccb_p cp);
void sym_xpt_async_bus_reset(hcb_p np);
void sym_xpt_async_sent_bdr(hcb_p np, int target);
int  sym_setup_data_and_start(hcb_p np, cam_scsiio_p csio, ccb_p cp);
void sym_log_bus_error(hcb_p np);
void sym_sniff_inquiry(hcb_p np, cam_ccb_p ccb, int resid);

#endif /* SYM_GLUE_H */
